//
// swupdate_locker.c
//

#include "swupdate_locker.h"

#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif
#include <sched.h>

#include <sys/eventfd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <poll.h>
#include <signal.h>
#include <unistd.h>

#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define LOCK_DIR	"/tmp/.swupdate_lock"
#define LOCK_FILE	LOCK_DIR "/pid"


struct ChildContext {
	bool	fatal;
	bool	result;
	int	childEvent;
};

#define STACK_SIZE	(8 * 1024)

static int	sChildTid;
static unsigned char	sStackArea[STACK_SIZE];
static struct ChildContext	sChildContext;


static bool
rewrite_proc_comm()
{
	bool	result = false;
	int	fd;
	const char*	name = "swupdate\n";

	fd = open("/proc/self/comm", O_WRONLY|O_TRUNC);
	if (fd >= 0) {
		(void)write(fd, name, strlen(name));
		(void)close(fd);
		result = true;
	}

	return result;
}

static void
write_pid(const char* path, const char* pid)
{
	int	fd = open(path, O_WRONLY|O_TRUNC|O_CREAT, 0664);

	if (fd >= 0) {
		(void)write(fd, pid, strlen(pid) + 1);
		(void)close(fd);
	}
}

static bool
try_lock()
{
	char	myPid[20+1+1];

	sprintf(myPid, "%d\n", getpid());

do_try:
	if (mkdir(LOCK_DIR, 0755) == 0) {
		write_pid(LOCK_FILE, myPid);
		return true;
	}
	if (access(LOCK_DIR, F_OK) == 0) {
		char	pid[20+1];
		int	fd;
		ssize_t	readLen;

		// there is a small window where directory exists but not pid file
		// cheat around it with a sleep...
		if (access(LOCK_FILE, F_OK) != 0) {
			(void)sleep(1);
		}

		fd = open(LOCK_FILE, O_RDONLY);
		if  (fd >= 0) {
			readLen = read(fd, pid, sizeof(pid));
			(void)close(fd);
		}
		if (fd < 0 || readLen <= 0) {
			(void)system("rm -rf " LOCK_DIR);
			goto do_try;
		}
		pid[readLen] = '\0';

		if (! is_live_process_of_swupdate(pid)) {
			(void)system("rm -rf " LOCK_DIR);
			goto do_try;
		}
	} else {
		// mkdir failed but lock does not exist, we could hove been
		// raced so try again once
		if (mkdir(LOCK_DIR, 0755) == 0) {
			write_pid(LOCK_FILE, myPid);
			return true;
		}
		if (access(LOCK_DIR, F_OK) != 0) {
			printf("Could not create install lock.\n");
		}
	}

	return false;
}

static void
signal_handler(int sig)
{
	(void)sig;
	// do nothing
}

static int
child_proc(void* arg)
{
	struct ChildContext*	context = (struct ChildContext*)arg;
	int64_t	evtVal = 1;

	context->fatal = false;
	if (rewrite_proc_comm()) {  // overwite /proc/<mypid>/comm to 'swupdate'
		context->result = try_lock();
	} else {
		context->fatal = true;
		context->result = false;
	}

	// notify to parent that the processing is done
	(void)write(context->childEvent, &evtVal, sizeof(evtVal));
	if (! context->result) {
		return 0;  // finish if the lock has not acquired
	}

	// wait the notification from parent
	(void)signal(SIGINT, signal_handler);
	(void)sleep((unsigned int)-1);

	// release the lock
	(void)system("rm -rf " LOCK_DIR);

	return 0;
}

static void
wait_child_and_cleanup(struct ChildContext* context)
{
	if (sChildTid == 0) {
		return;
	}
	(void)waitpid(sChildTid, NULL, __WCLONE);
	(void)close(context->childEvent);
	context->childEvent = 0;
	sChildTid = 0;
}

bool
acquire_swupdate_lock(bool* outAcquired)
{
	struct ChildContext*	context = &sChildContext;
	int	tid;
	int64_t	evtVal = 1;
	struct pollfd	pollArg;

	if (sChildTid != 0) {
		return false;  // the child has already started
	}

	// create the child process which like thread
	context->childEvent = eventfd(0, 0);
	if (context->childEvent == -1) {
		printf("failed to eventfd().\n");
		goto err;
	}
	tid = clone(
		child_proc, sStackArea + STACK_SIZE,
		CLONE_VM|CLONE_FILES, context);
	if (tid == -1) {
		printf("failed to clone.\n");
		goto err;
	}
	sChildTid = tid;

	// wait for the child's processing
	pollArg.fd = context->childEvent;
	pollArg.events = POLLIN;
	pollArg.revents = 0;
	if (poll(&pollArg, 1, 5000) <= 0) {
		printf("waiting child is timed out or failed.\n");
		context->result = false;
		context->fatal = true;
	} else {
		(void)read(context->childEvent, &evtVal, sizeof(evtVal)); 
	}
	if (! context->result) {
		wait_child_and_cleanup(context);
	}

	*outAcquired = context->result;
	return (! context->fatal);
err:
	if (context->childEvent != -1) {
		(void)close(context->childEvent);
	}
	return false;
}

void
release_swupdate_lock(void)
{
	struct ChildContext*	context = &sChildContext;

	if (sChildTid == 0) {
		return;  // the child has already stopped
	}

	// notify to the child that release the lock
	(void)kill(sChildTid, SIGINT);
	wait_child_and_cleanup(context);
}

bool
is_live_process_of_swupdate(const char* pid)
{
#define TASK_COMM_LEN   16  // from kernel's definition
        char    procPath[32];
        char    taskComm[TASK_COMM_LEN+1];
        int     fd;
        ssize_t readLen;

        sprintf(procPath, "/proc/%d/comm", atoi(pid));
        fd = open(procPath, O_RDONLY);
        if (fd < 0) {
                return false;
        }
        readLen = read(fd, taskComm, sizeof(taskComm) - 1);
        (void)close(fd);
        if (readLen <= 0) {
                return false;
        } else if (taskComm[readLen-1] == '\n') {
                --readLen;
                if (taskComm[readLen-1] == '\n') {
                        --readLen;
                }
        }
        taskComm[readLen] = '\0';

        return (0 == strcmp(taskComm, "swupdate"));
}

//
// End of File
//

